#!/usr/bin/python3
"""main file to see execution"""

from io import StringIO
from typing import List
import os
from pyjsonedit.tokenizer import tokenize
from pyjsonedit.tree import parse as tree_parse
from pyjsonedit.tree import JsonNode
from pyjsonedit.matcher import print_matched

def __get_tokens(json) -> List:
    tokens=[]
    if os.path.isfile(json):
        with open(json) as handle:
            tokens = list(tokenize(handle))
    else:
        with StringIO(json) as handle:
            tokens = list(tokenize(handle))
    return tokens

def string_to_tokens(json_str: str) -> List:
    """
    python3 -c 'from main import *; print( string_to_tokens("{}") );'
    """
    return __get_tokens(json_str)

def string_to_tree(json_str: str) -> JsonNode:
    """
    python3 -c 'from main import *; r=string_to_tree("{}"); print(r)'
    """
    tokens = __get_tokens(json_str)
    return tree_parse(tokens)

def string_match_mark(json, pattern, symbol='X', color=None):
    """mark part of matched json"""
    tokens = __get_tokens(json)
    node = tree_parse(tokens)
    return print_matched(json, node, pattern, symbol, color)

def cli_match_mark(pattern, json, symbol, color, callback=print):
    """cli method for masking matching parts of json"""

    tokens = []
    if os.path.isfile(json):
        with open(json) as handle:
            json = handle.read()

    with StringIO(json) as handle:
        tokens = list(tokenize(handle))

        node = tree_parse(tokens)
        ret = print_matched(json, node, pattern, symbol, color)
        callback(ret)



"""tokenize readable handle to tokens"""

def tokenize(handle):
    """read handle and turn it into tokens"""
    run = True
    pos = -1
    mem = ""
    while run:
        pos += 1
        char = handle.read(1)
        if not char:
            run = False
            break
        if char in ['[', ']', '{', '}', ",", ":"]:
            if mem.strip():
                yield("v", pos-len(mem), mem)
            mem = ""
            yield (char, pos)  # open
        elif char in ['"', "'"]:
            start_c = char
            pos_start = pos
            mem = ""
            while True:
                c_prev = char
                char = handle.read(1)
                pos += 1
                if not char:
                    run = False
                    break
                if char == start_c and c_prev != "\\":
                    break
                mem += char
            yield ('S', pos_start, mem)
            mem = ""
        else:
            mem += char



"""
this module alows to freely move throug JsonNode's
"""
import re
from pyjsonedit.tree import JsonNode

class MatchException(Exception):
    """failures in maching process"""
    def __eq__(self, other):
        return isinstance(other, MatchException) and \
            other.__str__() == self.__str__()

def _each_child_by_index(node:JsonNode, node_type:str, child_idx):
    if node.type != node_type:
        raise MatchException(f'unexpected node type: "{node.type}" expected "{node_type}"')
    if child_idx >= len(node.kids):
        raise MatchException(f'not enough kids {child_idx}')
    return node.kids[child_idx]

def _match_node(node:JsonNode, patterns, depth=0):

    try:
        if depth >= len(patterns) or not patterns[depth].strip():
            yield node
            return

        pattern = patterns[depth].strip()

        # array by index
        array_index = re.match(r'\[(\d+)\]',pattern)
        if array_index:
            child_idx = int(array_index.group(1))
            node = _each_child_by_index(node, 'array', child_idx)
            yield from _match_node( node, patterns, depth+1)
            return

        # dict by index
        dict_index  = re.match(r'\{(\d+)\}',pattern)
        if dict_index:
            child_idx = int(dict_index.group(1))
            node = _each_child_by_index(node, 'dict', child_idx)
            yield from _match_node( node, patterns, depth+1)
            return

        # raw
        if node.type == 'dict':
            found=False
            for k in node.kids:
                if pattern in ['*', k.name]:
                    found=True
                    yield from _match_node( k, patterns, depth+1)
            if found:
                return

        raise MatchException(f'pattern "{pattern}" not found')
    except MatchException as fail:
        yield fail

def match(root: JsonNode, pattern: str):
    """ start node matching"""
    patterns = pattern.strip().split('>')
    yield from _match_node(root, patterns)

def print_matched(json_str, node:JsonNode, pattern:str, mark_symbol='X', color=False):
    """generates marked characters matching given 'pattern'"""

    found = []
    for i in match(node, pattern):
        if isinstance(i, MatchException):
            raise i
        found.append( (i.start,i.end))

    ret = ""
    for i,char in enumerate(json_str):
        mark = False
        for start,end in found:
            if start <= i < end:
                mark = True
                break
        if color:
            color = '\033[91m' if mark else '\033[92m'

        if mark:
            char=mark_symbol

        ret += f'{color}{char}\033[0m' if color else char

    return ret
